# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lacss.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0blacss.proto\x12\x0ftrackmate.lacss\"\xba\x01\n\x05Image\x12\x0e\n\x06height\x18\x01 \x01(\x04\x12\r\n\x05width\x18\x02 \x01(\x04\x12\x0f\n\x07\x63hannel\x18\x03 \x01(\x04\x12%\n\x05\x64type\x18\x04 \x01(\x0e\x32\x16.trackmate.lacss.DType\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x12\r\n\x05\x64\x65pth\x18\x06 \x01(\x04\x12\x13\n\x0bvoxel_dim_x\x18\x07 \x01(\x02\x12\x13\n\x0bvoxel_dim_y\x18\x08 \x01(\x02\x12\x13\n\x0bvoxel_dim_z\x18\t \x01(\x02\"\xcd\x01\n\x08Settings\x12\x15\n\rmin_cell_area\x18\x01 \x01(\x02\x12\x1b\n\x13remove_out_of_bound\x18\x02 \x01(\x08\x12\x0f\n\x07scaling\x18\x03 \x01(\x02\x12\x0f\n\x07nms_iou\x18\x04 \x01(\x02\x12\x1b\n\x13\x64\x65tection_threshold\x18\x05 \x01(\x02\x12\x1e\n\x16segmentation_threshold\x18\x06 \x01(\x02\x12\x16\n\x0ereturn_polygon\x18\x07 \x01(\x08\x12\x16\n\x0e\x63\x65ll_size_hint\x18\x08 \x01(\x02\"[\n\x05Input\x12+\n\x08settings\x18\x01 \x01(\x0b\x32\x19.trackmate.lacss.Settings\x12%\n\x05image\x18\x02 \x01(\x0b\x32\x16.trackmate.lacss.Image\"(\n\x05Point\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"@\n\x07Polygon\x12\r\n\x05score\x18\x01 \x01(\x02\x12&\n\x06points\x18\x02 \x03(\x0b\x32\x16.trackmate.lacss.Point\"\x84\x01\n\x04Mesh\x12\r\n\x05score\x18\x01 \x01(\x02\x12%\n\x05verts\x18\x02 \x03(\x0b\x32\x16.trackmate.lacss.Point\x12\r\n\x05\x66\x61\x63\x65s\x18\x03 \x03(\x04\x12\'\n\x07normals\x18\x04 \x03(\x0b\x32\x16.trackmate.lacss.Point\x12\x0e\n\x06values\x18\x05 \x03(\x02\"`\n\x03Roi\x12+\n\x07polygon\x18\x01 \x01(\x0b\x32\x18.trackmate.lacss.PolygonH\x00\x12%\n\x04mesh\x18\x02 \x01(\x0b\x32\x15.trackmate.lacss.MeshH\x00\x42\x05\n\x03roi\"-\n\x07Results\x12\"\n\x04rois\x18\x01 \x03(\x0b\x32\x14.trackmate.lacss.Roi*\x14\n\x05\x44Type\x12\x0b\n\x07\x46LOAT32\x10\x00\x32K\n\x05Lacss\x12\x42\n\x0cRunDetection\x12\x16.trackmate.lacss.Input\x1a\x18.trackmate.lacss.Results\"\x00\x42\x30\n$fiji.plugin.trackmate.detector.lacssB\x08LacssMsgb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lacss_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n$fiji.plugin.trackmate.detector.lacssB\010LacssMsg'
  _globals['_DTYPE']._serialized_start=910
  _globals['_DTYPE']._serialized_end=930
  _globals['_IMAGE']._serialized_start=33
  _globals['_IMAGE']._serialized_end=219
  _globals['_SETTINGS']._serialized_start=222
  _globals['_SETTINGS']._serialized_end=427
  _globals['_INPUT']._serialized_start=429
  _globals['_INPUT']._serialized_end=520
  _globals['_POINT']._serialized_start=522
  _globals['_POINT']._serialized_end=562
  _globals['_POLYGON']._serialized_start=564
  _globals['_POLYGON']._serialized_end=628
  _globals['_MESH']._serialized_start=631
  _globals['_MESH']._serialized_end=763
  _globals['_ROI']._serialized_start=765
  _globals['_ROI']._serialized_end=861
  _globals['_RESULTS']._serialized_start=863
  _globals['_RESULTS']._serialized_end=908
  _globals['_LACSS']._serialized_start=932
  _globals['_LACSS']._serialized_end=1007
# @@protoc_insertion_point(module_scope)



import logging
from concurrent import futures

from pathlib import Path
import threading

import grpc
import jax
import numpy as np
import typer

from .proto import lacss_pb2, lacss_pb2_grpc

_AUTH_HEADER_KEY = "authorization"

app = typer.Typer(pretty_exceptions_enable=False)

_MAX_MSG_SIZE=1024*1024*128
_TARGET_CELL_SIZE=32

def process_input(msg):
    image = msg.image
    settings = msg.settings

    dim_x = image.voxel_dim_x or 1.0
    dim_y = image.voxel_dim_y or 1.0
    dim_z = image.voxel_dim_z or 1.0

    if settings.cell_size_hint != 0:
        scale_x = _TARGET_CELL_SIZE / settings.cell_size_hint * dim_x
        scale_y = _TARGET_CELL_SIZE / settings.cell_size_hint * dim_y
        scale_z = _TARGET_CELL_SIZE / settings.cell_size_hint * dim_z
    else:
        scale_x = scale_y = scale_z = settings.scaling or 1.0
        z_ratio = dim_z / dim_x
        scale_z = scale_z * z_ratio

    logging.debug(f"Requested rescaling factor is {(scale_z, scale_y, scale_x)}")

    np_img = np.frombuffer(image.data, dtype=">f4").astype("float32")
    if image.depth == 0 or image.depth == 1:
        np_img = np_img.reshape(image.height, image.width, image.channel)
        shape_hint = (
            int(scale_y * image.height + .5),
            int(scale_x * image.width + .5),
        )
    else:
        np_img = np_img.reshape(image.depth, image.height, image.width, image.channel)
        shape_hint = (
            int(scale_z * image.depth +.5),
            int(scale_y * image.height + .5),
            int(scale_x * image.width + .5),
        )

    return np_img, shape_hint


def process_result(preds, is3d):
    msg = lacss_pb2.Results()
    if not is3d:
        for polygon, score in zip(preds["pred_contours"], preds["pred_scores"]):
            if len(polygon) == 0:
                continue
            roi = lacss_pb2.Roi()
            roi.polygon.score = score
            for p in polygon:
                roi.polygon.points.append(lacss_pb2.Point(
                    x=p[0],
                    y=p[1]
                ))
            msg.rois.append(roi)

    else:
        for mesh, score in zip(preds["pred_contours"], preds["pred_scores"]):
            roi = lacss_pb2.Roi()
            roi.mesh.score = score
            for v in mesh['verts']:
                point = lacss_pb2.Point(
                    z = v[0], y = v[1], x = v[2]
                )
                roi.mesh.verts.append(point)
            for f in mesh['faces']:
                roi.mesh.faces.append(f[0])
                roi.mesh.faces.append(f[1])
                roi.mesh.faces.append(f[2])

            msg.rois.append(roi)

    return msg


class TokenValidationInterceptor(grpc.ServerInterceptor):
    def __init__(self, token):
        def abort(ignored_request, context):
            context.abort(grpc.StatusCode.UNAUTHENTICATED, "Invalid token signature")

        self._abort_handler = grpc.unary_unary_rpc_method_handler(abort)
        self.token = token

    def intercept_service(self, continuation, handler_call_details):
        expected_metadata = (_AUTH_HEADER_KEY, f"Bearer {self.token}")
        if self.token is None or expected_metadata in handler_call_details.invocation_metadata:
            return continuation(handler_call_details)
        else:
            return self._abort_handler


class LacssServicer(lacss_pb2_grpc.LacssServicer):
    def __init__(self, model):
        self.model = model
        self._lock = threading.RLock()

    def RunDetection(self, request, context):
        with self._lock:
            img, shape_hint = process_input(request)
            settings = request.settings
            is3d = img.ndim == 4

            is_img_too_big = is3d and max(shape_hint) > 256
            is_img_too_big |= (not is3d) and max(shape_hint) > 1088
            if is_img_too_big:
                context.abort(grpc.StatusCode.INVALID_ARGUMENT, "image size exeeds limit")

            logging.debug(f"received image {img.shape}")

            # dont' reshape image if the request is almost same as the orginal 
            rel_diff = np.abs(np.array(shape_hint) / img.shape[:-1] - 1)
            if (rel_diff < 0.1).all():
                shape_hint = None

            try:
                preds = self.model.predict(
                    img,
                    reshape_to = shape_hint,
                    min_area=settings.min_cell_area,
                    nms_iou=settings.nms_iou,
                    score_threshold=settings.detection_threshold,
                    segmentation_threshold=settings.segmentation_threshold,
                    output_type="contour",
                )

                results = process_result(preds, is3d)

                logging.debug(f"Reply with message of size {results.ByteSize()}")

                return results

            except Exception as e:

                logging.error(f"prediction failed: {repr(e)}")

                context.abort(grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}")


def get_predictor(modelpath):
    from .predict import Predictor

    model = Predictor(modelpath)

    logging.info(f"lacss_server: loaded model from {modelpath}")

    model.module.detector.max_output = 512  # FIXME good default?
    model.module.detector.min_score = 0.05

    logging.debug(f"lacss_server: precompile for most common shapes")

    _ = model.predict(np.ones([512,512,3]))
    _ = model.predict(np.ones([1088,1088,3]))
    _ = model.predict(np.ones([256,256,256,3]))

    return model


def show_urls():
    from .predict import model_urls
    
    print("Pretrained model files:")
    print("==============================")
    for k, v in model_urls.items():
        print(f"{k}: {v}")
    print()


@app.command()
def main(
    modelpath: Path|None = None,
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool|None = None,
    debug: bool = False,
    compression: bool = True,
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    if modelpath is None:
        show_urls()
        return

    print ("server starting ...")

    model = get_predictor(modelpath)

    logging.info(f"lacss_server: default backend is {jax.default_backend()}")

    if jax.default_backend() == "cpu":
        logging.warning(
            f"lacss_server: WARNING: No GPU configuration. This might be very slow ..."
        )

    if token is None:
        token = not local
    if token:
        import secrets

        token = secrets.token_urlsafe(64)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token}")
        print("=======================================================================")
        print()
    else:
        token = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip if compression else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token),),
        options=(("grpc.max_receive_message_length", _MAX_MSG_SIZE),),
    )

    lacss_pb2_grpc.add_LacssServicer_to_server(LacssServicer(model), server)

    if local:
        server.add_secure_port(f"localhost:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logging.info(f"lacss_server: listening on port {port}")

    print ("server starting ... ready")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    app()
